'''Generate card-style stimuli on the fly.

Card-stimuli can contain patterns and colors and different cards can be
quickly and efficiently swapped to other cards.

It is well suited for presenting different kind of non-moving
patters and images.

Use somthing else for high-framerate moving stimuli or long videos.
'''
import math
import random

from PIL import Image, ImageDraw

import devjoni.guibase as gb


CARD_WIDTH = 200
CARD_HEIGHT = 200


def _calc_variations(N):
    '''Calculate the pie binary variations
    '''
    
    if N == 4:
        # Alice style variations (rotations eliminated)
        variations = ["0000", "0001", "0011", "1010", "0111", "1111"]
        return variations

    # General case but does not check rotationally 
    # similar variations

    variations = []
    variation = '0'*N
    while True:
        # Advance variation by one and quit when done
        variation = bin(int(variation,2)+1)[2:].zfill(N)
        if len(variation) > N:
            break
        variations.append(variation)
    return variations

def _randomize_rotation(variation):
    # Cut index
    index = random.randint(0, len(variation)-1)
    return variation[index:] + variation [:index]

def _draw_pie(image, x0, y0, x1, y1, variation):
    '''Draw a pie on the Pillow image

    Arguments
    ---------
    image : Image object
        Pillow image
    x0, y1, x1, y1 : int
        Coordinates of the area filled with the pie
    variation : string
        Variation of the pie, for example "0101"
    '''
    ctx = ImageDraw.Draw(image)

    variation = _randomize_rotation(variation)

    opening = 360 / len(variation)
    
    for i_boolean, boolean in enumerate(variation):
        if boolean == '1':
            ctx.pieslice(
                    [(x0,y0),(x1,y1)],
                    i_boolean*opening, (i_boolean+1)*opening,
                    fill=(255,255,255))

    w = x1-x0
    h = y1-y0

    cp = [int(w/2+x0),int(h/2+y0)]
    r = int(min(w/2,h/2))
    ctx.circle(cp,r)



def create_onepie_images(N, width=CARD_WIDTH, height=CARD_HEIGHT,
                         seed=None):
    '''Create a image with one pie

    Attributes
    ----------
    N : int
        The amount of slices in the pattern
    '''
    setseed(seed)

    images = []

    for variation in _calc_variations(N):        
        image = Image.new('RGB', (width, height))
        _draw_pie(image, 0, 0, width, height, variation)
        images.append(image)

    setseed(None)

    return images


def setseed(state):
    random.seed(state)



def create_multipie_images(N, M, right='1010', width=CARD_WIDTH, height=CARD_HEIGHT, seed=None):
    '''Create a image with one pie

    Attributes
    ----------
    N : int
        The amount of slices in the pattern
    M : int
        The amount of patterns
    '''
    images = []

    setseed(seed)

    right = right.zfill(M)[0:M]
    
    variations = _calc_variations(N)
    if right in variations:
        variations.remove(right)

    mrot = 360/M
    wpie = min(width, height)/M


    for i in range(12):
        
        right_rot = random.randint(0,M-1)

        image = Image.new('RGB', (width, height))
        
        for irot in range(M):
            cp = [
                    math.cos(math.radians(irot*mrot))*(
                        width/2-wpie)+ width/2,
                    math.sin(math.radians(irot*mrot))*(
                        height/2-wpie) + height/2]

            if irot == right_rot:
                variation = right
            else:
                variation = random.choice(variations)

            _draw_pie(
                    image,
                    cp[0]-wpie/2, cp[1]-wpie/2,
                    cp[0]+wpie/2, cp[1]+wpie/2,
                    variation)

        images.append(image)
    
    setseed(None)


    return images




class CardWidget(gb.FrameWidget):
    '''A stimulus

    Load method is pretty slow, Hide and Show fast.
    '''
    def __init__(self, parent, width=CARD_WIDTH, height=CARD_HEIGHT):
        super().__init__(parent)
        
        image = gb.ImageImage(None, width, height)
        self.widget = gb.ImageWidget(self, image)
        self.widget.grid()

        self.set(bg="black")
        self.widget.set(bg="black")


    def load(self, image):
        '''Load image

        image : Image
            Pillow image object
        '''
        self.widget.image.set_from_rgb(image)


class CardStimWidget(gb.FrameWidget):
    '''Generate different simulus cards

    Attributes
    ----------
    next_card_callback : None or callable
        Function or method to be called every time the card is changed
    '''
    def __init__(self, parent, width=CARD_WIDTH, height=CARD_HEIGHT,
                 make_nextbutton=True):
        super().__init__(parent)

        self.width = width
        self.height = height

        self.cards = []
        self.current_card = None
        
        if make_nextbutton:
            self.b_next = gb.ButtonWidget(self, 'Next', command=self.next_card)
            self.b_next.grid(row=0, column=0)
        
        self.set(bg="black")

        self.next_card_callback = None

    def next_card(self):
        if not self.cards:
            return

        if self.current_card is None:
            index = 0
        else:
            index = self.cards.index(self.current_card)
            self.current_card.grid_remove()
            index+=1

        if index >= len(self.cards)-1:
            index = 0
       
        self.current_card = self.cards[index]
        self.current_card.grid(row=1, column=0)

        if callable(self.next_card_callback):
            self.next_card_callback()

    def create_card(self, image):
        '''Manually create and add a card from the given image

        image : Image object
            Pillow image
        '''
        h = image.height
        w = image.width

        converted = []
        _converted = list(image.getdata())
        for i_row in range(image.height):
            row = _converted[i_row*w:(i_row+1)*w]
            converted.append(row)

        card = CardWidget(self, self.width, self.height)
        card.load(converted)

        self.cards.append(card)

    def clear_cards(self):
        self.cards = []

    def create_onepie_cards(self, N=4, seed=None):
        '''Change to onepie cards
        '''
        self.clear_cards()

        images = create_onepie_images(
                N, width=self.width, height=self.height,
                seed=seed)
        for image in images:
            self.create_card(image)
        self.current_card = None


    def create_multipie_cards(self, N=4, M=4, seed=None):

        self.clear_cards()

        images = create_multipie_images(
                N, M, width=self.width, height=self.height,
                seed=seed)
        for image in images:
            self.create_card(image)

        self.current_card = None


def main():
    
    window = gb.MainWindow()
    window.title = f'Test stimulus'

    
    view = CardStimWidget(window)
    view.grid()

    view.create_multipie_cards()

    window.run()

if __name__ == "__main__":
    main()



'''Arena GUI program
'''

import sys
import atexit
import platform
import random

import devjoni.guibase as gb
from devjoni.hosguibase.video import VideoWidget

from .arenalib import Arena
from .cardstimgen import CardStimWidget

from .version import __version__

IMAGE_UPDATE_INTERVAL = 10 # ms


class MovementView(gb.FrameWidget):
    '''Control the arena lift up and down
    '''

    def __init__(self, parent, arena):
        super().__init__(parent)

        self.arena = arena

        self.pos = gb.TextWidget(self, '')
        self.pos.set(text=f'{self.arena.pos}')
        self.pos.grid(row=0, column=0)

        self.move_buttons = []

        for i, N in enumerate([10,3,1,-1,-3,-10]):
            b = gb.ButtonWidget(
                    self, f'Move {N}',
                    lambda N=N: self.move(N)
                    )
            b.grid(row=i+1, column=0)
            self.move_buttons.append(b)

        align = gb.ButtonWidget(self, 'End-align', self.do_align)
        align.grid(row=i+2, column=0)
        

    def move(self, N_steps):
        self.arena.move_platform(N_steps)
        self.pos.set(text=f'{self.arena.pos} ')

    def do_align(self):
        self.arena.step_end_align()
        self.pos.set(text=f'{self.arena.pos}')


class LightView(gb.FrameWidget):
    '''Control the arena lights
    '''
    def __init__(self, parent, arena):
        super().__init__(parent)

        self.arena = arena
        
        self.led_buttons = []
        for i_led in range(self.arena.get_N_leds()):
            b = gb.ButtonWidget(
                    self, f'LED {i_led+1}',
                    command=lambda i=i_led: self.toggle(i))
            b.grid(row=i_led, column=0)
            self.led_buttons.append(b)
    
    def toggle(self, i_led):
        state = self.arena.get_led(i_led)
        state = not state

        self.arena.set_led(i_led, state)
        
        if state:
            self.led_buttons[i_led].set(bg='green')
        else:
            self.led_buttons[i_led].set(bg='gray')

class StimView(gb.FrameWidget):
    '''Control the stimulus presentation
    '''
    def __init__(self, parent):

        super().__init__(parent)

        self.b_generate = gb.ButtonWidget(
                self, 'Generate cards',
                command=self.generate_cards)
        self.b_generate.grid(row=0, column=0)
        
        self.b_open = gb.ButtonWidget(
                self, 'Open in window',
                command=self.open_window)
        self.b_open.grid(row=2, column=0)
        
        self.preview = CardStimWidget(self, 100, 100)
        self.preview.grid(row=1, column=0)

        self.view = None

    def generate_cards(self):
        seed = random.random()

        self.preview.create_multipie_cards(seed=seed)
        self.preview.next_card()
    
        if self.view:
            self.view[1].create_multipie_cards(seed=seed)
            self.view[1].next_card()

    def open_window(self):
        
        if self.view is not None:
            toplevel, view = self.view
            view.destroy()
            toplevel.destroy()
            self.preview.next_card_callback = None

        root = self.get_root()    
        toplevel = gb.MainWindow(parent=root)
        
        view = CardStimWidget(toplevel, 400, 400, make_nextbutton=False)
        #view.b_next.destroy()
        self.preview.next_card_callback = view.next_card
        view.grid()

        self.view = [toplevel, view]
        

class CameraControlView(gb.FrameWidget):
    '''Camera controls like start imaging, recording etc.

    Attributes
    ----------
    camera_view : obj or None
        A CameraView Widget to be controlled
    '''
    def __init__(self, parent, camera_view=None):
        super().__init__(parent)

        self.camera_view = camera_view

        self.play = gb.ButtonWidget(self, text='Play', command=self.play)
        self.play.grid(row=0, column=0)
        self.stop = gb.ButtonWidget(self, text='Stop', command=self.stop)
        self.stop.grid(row=0, column=1)
        
        self.change = gb.ButtonWidget(
                self, text='Change camera', command=self.next_camera)
        self.change.grid(row=0, column=2)
   
        self.record = gb.ButtonWidget(
                self, text='Record', command=self.record)
        self.record.grid(row=0, column=3)
        
        self.record_details = gb.TextWidget(
                self, text='')
        self.record_details.grid(row=0, column=4)
  

    def play(self):
        self.camera_view.play()

    def stop(self):
        self.camera_view.stop()
    
    def record(self):
        self.camera_view.record()
        if self.camera_view.do_record:
            self.record.set(bg="red")
            fn = self.camera_view.record_fn
            fps = self.camera_view.record_fps
            self.record_details.set(
                    text=f'{fn} @ {fps}'
                    )
        else:
            self.record.set(bg="gray")
            self.record_details.set(text='Recording finished')

    def next_camera(self):
        self.camera_view.next_camera()
        if self.camera_view.do_record:
            self.record.set(bg="gray")
            self.record_details.set(text='Recording finished')


class CameraView(gb.FrameWidget):
    '''Camera view using the opencv cameralib.
    '''
    def __init__(self, parent):
        super().__init__(parent)

        cams = detect_cameras()
        self.camera = MultiprocessCamera(cams[0])

        
        self.image = gb.ImageImage(None, 200,200)
        self.canvas = gb.ImageWidget(self, self.image)

        self.canvas.grid(0,0, sticky='NSWE')

        self._is_playing = False

        atexit.register(self.stop)

    def play(self):
        if self._is_playing:
            return
        self._is_playing = True
        self.camera.open()
        self.after(IMAGE_UPDATE_INTERVAL, self.tick)
        
    def stop(self):
        self._is_playing = False
        self.camera.close()


    def tick(self):
        if not self.camera.is_open():
            return
        im = self.camera.get_frame()
        if im is not None:
            print(im)
            self.canvas.image.set_from_rgb(im)
        
        if self._is_playing:
            self.after(IMAGE_UPDATE_INTERVAL, self.tick)


    def next_camera(self):
        pass


class FastCameraView(gb.FrameWidget):
    def __init__(self, parent):
        super().__init__(parent)
        
        self.video = VideoWidget(self)
        self.video.grid()
        self.video.relative_size=0.75
        self.video.tcoder.source_type = "camera"

        self.ffmpeg_line = ''

        self.do_record = False
        self.record_fn = 'test.mp4'
        self.record_fps = 100

    
        cameras = self.video.tcoder.detect_cameras()
        if cameras:
            self.video.source = cameras[0]


    def play(self):
        if self.do_record:
            self.video.tcoder.fps = self.record_fps
            self.video.tcoder.set_video_output(
                    self.record_fn,
                    opts=[
                        '-c:v', 'libx264',
                        '-preset', 'ultrafast',
                        '-crf', '28',
                        ])
            
            # Enforce that the webcam is being read fast
            # Arducam UC-844
            system = platform.system()
            if system == 'Linux':
                opts = ['-f', 'v4l2', '-framerate', 100,
                        '-video_size', '1280x800', '-input_format',
                        'mjpeg']
            else:
                opts = ['-f', 'dshow', '-video_size', '1280x800',
                        '-rtbufsize', '200M',
                        '-framerate', 100, '-vcodec', 'mjpeg']
            self.video.tcoder.source_opts = opts
        else:
            self.video.tcoder.fps = self.video.fps
            self.video.tcoder.set_video_output(None)
            self.video.tcoder.source_opts = None
        self.video.start()

    def stop(self, stop_record=True):
        if self.do_record and stop_record:
            self.do_record = False
        self.video.stop()
    
    def record(self):
        if self.do_record:
            self.do_record = False
        else:
            self.do_record = True
        self.stop(stop_record=False)
        self.play()
    
    def next_camera(self):
        
        # Always stop recording when changing the camera
        if self.do_record:
            self.do_record=False
            self.stop()

        cameras = self.video.tcoder.detect_cameras()
        if not cameras:
            return
        
        if self.video.source not in cameras:
            camera = cameras[0]
        else:
            index = cameras.index(self.video.source)
            index += 1

            if index >= len(cameras):
                index = 0
            camera = cameras[index]
        
        print(camera)
        self.stop()
        self.video.source = camera
        self.play()


class TotalView(gb.FrameWidget):
    def __init__(self, parent):
        super().__init__(parent)

        # Camera side

        camerabox = gb.FrameWidget(self)
        camerabox.grid(row=0, column=1, rowspan=2)
        
        control = CameraControlView(camerabox)
        control.grid(row=0, column=0, sticky='')

        camera = FastCameraView(camerabox)
        camera.grid(row=1, column=0)

        control.camera_view = camera

        # Motion control side
        
        try:
            arena = Arena()
        except:
            arena = Arena(fake_serial=True)

        controlbox = gb.FrameWidget(self)
        controlbox.grid(row=0, column=0)
 
        movement = MovementView(controlbox, arena)
        movement.grid(row=0, column=0)
        
        light = LightView(controlbox, arena)
        light.grid(row=0, column=1)

        # Stimulus

        stim = StimView(self)
        stim.grid(row=1, column=0)


def main():

    window = gb.MainWindow()
    window.title = f'Arena Program - v{__version__}'

    view = TotalView(window)
    view.grid()


    window.run()

if __name__ == "__main__":
    main()
    #cProfile.run('main()')



'''Python library for the controlling the arena9
'''

import os
import platform

import serial


class FakeSerial:
    def write(self, message):
        return f'fakeserial: {message}'
    def readline(self):
        return ''

def _say(ser, message):
    ser.write(message.encode('ASCII'))
    return ser.readline()

def detect_controller_devices():
    devs = []

    system = platform.system()
    if system == "Linux":
        devs = [os.path.join('/dev', fn) for fn in os.listdir(
            '/dev') if fn.startswith('ttyACM')]
    elif system== "Windows":
        coms = [f'COM{i}' for i in range(256)]
        for com in coms:
            try:
                ser = serial.Serial(com, 9600)
                ser.close()
                devs.append(com)
            except:
                pass
    else:
        raise NotImplementedError(f'Unkown system: {system}')
    return devs


def toggle_led(ser, i_led, value):
    leds_off = ['a', 'b', 'c', 'd', 'e', 'f']
    leds_on = ['A', 'B', 'C', 'D', 'E', 'F']
    
    if not 0 <= i_led <= len(leds_off)-1:
        return ValueError("i_led has to be between 0 and 3")

    if value > 0:
        message = leds_on[i_led]
    else:
        message = leds_off[i_led]
    return _say(ser, message)


def move_platform_up(ser, N_steps):
    N_steps = int(round(N_steps))
    return _say(ser, str('r'*N_steps))

def move_platform_down(ser, N_steps):
    N_steps = int(round(N_steps))
    return _say(ser, str('l'*N_steps))

def step_end_align(ser):
    return _say(ser, 'x')


class Arena:
    '''Control the Alice Arena9

    Attributes
    ----------
    ser : obj
        The serial library object
    '''
    def __init__(self, device=None, fake_serial=None):
        
        self.pos = 0
        self.led_states = {}
        
        if fake_serial:
            self.ser = FakeSerial()
            return

        if device is None:
            devs = detect_controller_devices()
            if not devs:
                raise RuntimeError("Could not detect the arena")
            device = devs[0]

        self.ser = serial.Serial(device, 9600)
        
        

    def set_led(self, i_led, value):
        '''Set an LED on or off
        '''
        self.led_states[i_led] = value
        return toggle_led(self.ser, i_led, value)

    def get_led(self, i_led):
        '''Returns True if the LED is on, otherwise False
        '''
        return bool(self.led_states.get(i_led, False))


    def get_N_leds(self):
        '''Returns the total amount of leds
        '''
        # In the future, this will be checked from the device
        # likely but the first shipped proto has 6 leds :)
        return 6

    def move_platform(self, N_steps):
        '''Move the platform up or down

        Arguments
        N_steps : int
            Positive for rising steps, moving the platform up.
            Negative for going down. Zero doesn't do anything.
        '''

        self.pos += N_steps

        if N_steps > 0: 
            return move_platform_up(self.ser, N_steps)
        elif N_steps < 0:
            return move_platform_down(self.ser, -N_steps)
    
    def step_end_align(self):
        '''Do the end align with little torque
        '''
        self.pos = 0
        return step_end_align(self.ser)


def main():

    arena = Arena()
    ser = arena.ser

    while True:
        
        message = input('LETTER >> ')
        print( _say(ser, message)) 


if __name__ == "__main__":
    main()



''' Cameralib - Camera reading using opencv
'''

import cv2
import multiprocessing

from devjoni.hosguibase.imagefuncs import rgb2hex

def detect_cameras():
    cameras = []
    for i in range(0,5):
        try:
            cam = cv2.VideoCapture(i)
            cameras.append(i)
        except:
            break
    return cameras


class Camera:
    def __init__(self, i_camera):
        self.i_camera = i_camera
        self.cam = cv2.VideoCapture(i_camera)
        self._is_open = False

    def get_frame(self):
        if not self._is_open:
            raise RuntimeError('Camera has not been opened')

        self.cam.grab()
        ret , im = self.cam.retrieve()
        #self.cam.release()
        im = cv2.resize(im, (200,200))
        a = rgb2hex(im, N_jobs=0)
        return im


    def open(self):
        self.cam.open(self.i_camera)
        self._is_open = True

    def close(self):
        self.cam.release()
        self._is_open = False

    def is_open(self):
        return self._is_open

class MultiprocessCamera:
    '''Start the camera in its own process
    '''

    def __init__(self, i_camera):
        self.i_camera = i_camera
        self.p = None

    def _run(self, q1, q2, i_camera):

        camera = Camera(i_camera)
        camera.open()

        while True:

            im = camera.get_frame()
            q1.put(im)

            if not q2.empty():
                camera.close()
                return


    def get_frame(self):
        if self.p is None:
            raise RuntimeError('Camera has not been opened')

        while True:
            if not self.q1.empty():
                break
        
        while True:
            im = self.q1.get()
            if self.q1.empty():
                break

        return im

    def open(self):
        q1 = multiprocessing.Queue()
        q2 = multiprocessing.Queue()
        self.p = multiprocessing.Process(target=self._run, args=[q1,q2, self.i_camera])
        self.q1 = q1
        self.q2 = q2
        self.p.start()


    def close(self):
        if self.p is None:
            return

        self.q2.put('')
        while True:
            try:
                self.p.join(1)
                break
            except:
                print('join failed')
            print(f'Waiting to join')
        self.p = None


    def is_open(self):
        return not (self.p is None)



